#!/usr/bin/perl

use strict;
use warnings;
use Time::Local 'timelocal_nocheck';
use File::Basename;

my ($line, %EVENT_MAP_HASH, %EVENT_TYPE_HASH, $sample_line, $file_name, $line_num, $prev_original_line);
my ($timestamp, $rest, $year, $month, $day, $hour, $minute, $second, $msec, $exception_block);
my ($event_id, $template, $prev_line, %TMP_HASH, $name, $value, $hash, $count, %EXCEPTION_LINE_HASH);
my (@tmp_array, $index, $max_event_id, $tmp_string, %NEW_EVENTS, %EVENT_TYPE_HASH2, %ORIGINAL_LINE_HASH);
my ($task_id, $job_id, $blk_id, $attempt_id, $output_file, %TMP_HASH2, $word, $jvm_id);
my ($name1, $value1, $hash1, $count1, %TIME_HASH, $prev_time, $EVENT_TYPE_HASH2, $original_line);
my ($prev_template, $prev_event_id, $prev_sample_line, $has_new_events, $header, $prev_header);

# read from the event map file
$max_event_id = -1;
$prev_template = "";
$prev_event_id = "";
$prev_sample_line = "";
$has_new_events = 0;

#$file_name = "hadoop-ian-jobtracker-sail213.log";
#$file_name = "test.txt";
$file_name = $ARGV[0];
open LOG_INPUT, "$file_name" or die "Cannot open $file_name: $!\n";
$line_num = 0;
$prev_line = "";
$prev_header = "";
$exception_block = 0; 
$prev_time = 0;
%EXCEPTION_LINE_HASH = ();
$prev_original_line = "";

while (<LOG_INPUT>) {
	$line = $_;
	chomp $line;
	$original_line = $line;

	# step 1. check if it is actually a log line not just the stack trace
	if ($line =~ /^(\w)(\/[\|\w\s\*\.-]+)+\s*\(\s*\d+\s*\):/) {
		# Andriod logcat output
		# I/ActivityManager(  386): Start proc com.android2.calculator3 for activity com.android2.calculator3/.Calculator: pid=29078 uid=10062 gids={50062, 1028}
		# D/VLC     (16435): core input: Creating an input for '(null)'
		# D/VLC/JNI/main(16435): Using deblocking level 3
		# W/System.err(29078): java.io.FileNotFoundException: /data/data/com.android2.calculator3/files/calculator.data: open failed: ENOENT (No such file or directory)
		if ($line =~ /^([^:]+)\(\s*\d+\s*\):\s+(.*)$/) {
			$header = $1 . "(TID): ";
			$rest = $2;

			# we need to decide whether it's part of the exception stack or not
			# E/GeckoAppShell( 7934): java.io.FileNotFoundException: /dev/cpuctl/tasks: open failed: EACCES (Permission denied)
			# E/GeckoAppShell( 7934): 	at libcore.io.IoBridge.open(IoBridge.java:416)
			# E/GeckoAppShell( 7934): 	at java.io.FileOutputStream.<init>(FileOutputStream.java:88)
			# E/GeckoAppShell( 7934): 	at java.io.FileOutputStream.<init>(FileOutputStream.java:73)
			# E/GeckoAppShell( 7934): 	at org.mozilla.gecko.GeckoAppShell$5.callback(GeckoAppShell.java:1109)
			# E/GeckoAppShell( 7934): 	at org.mozilla.gecko.GeckoAppShell.EnumerateGeckoProcesses(GeckoAppShell.java:1215)
			# E/GeckoAppShell( 7934): 	at org.mozilla.gecko.GeckoAppShell.putChildInBackground(GeckoAppShell.java:1120)
			# E/GeckoAppShell( 7934): 	at org.mozilla.gecko.GeckoApp.onStop(GeckoApp.java:391)
			# E/GeckoAppShell( 7934): Caused by: libcore.io.ErrnoException: open failed: EACCES (Permission denied)
			# E/GeckoAppShell( 7934): 	at libcore.io.Posix.open(Native Method)
			# E/GeckoAppShell( 7934): 	at libcore.io.BlockGuardOs.open(BlockGuardOs.java:110)
			# E/GeckoAppShell( 7934): 	at libcore.io.IoBridge.open(IoBridge.java:400)
			# E/GeckoAppShell( 7934): 	... 24 more

			if ($exception_block == 0) {
				# print STDERR "1234 $rest\n";

				# prev_line is normal log line
				if ($prev_header ne "") {
					$TMP_HASH{$line_num} = $prev_header . $prev_line;
					$ORIGINAL_LINE_HASH{$line_num} = $prev_original_line;
				} 

				# when we are not in the exception block in the previous lines
				if ($rest =~ /^[\w\.]*\.\w+Exception:/) {
					# we are in an exception block!
					$exception_block = 1;
					$prev_line = $rest;
				} else {
					# normal line
					$prev_line = $rest;
					$prev_header = $header;
					$prev_original_line = $original_line;
					$exception_block = 0;
				}
			} else {
				# print STDERR "5678 $rest\n";
				if ($rest =~ /^at / or $rest =~ /^Caused by:/ or $rest =~ /\.\.\. \d+ more/) {
					# continue in the exception block
					$prev_line = $prev_line . "\t\t\t\t\t" . $rest;
					$prev_header = $header;
					$exception_block = 1;
				} else {
					# no longer in the exception block, we are done
					$TMP_HASH{$line_num} = $prev_header . $prev_line;
					$EXCEPTION_LINE_HASH{$line_num}++;
					$prev_line = $rest;
					$prev_header = $header;
					$prev_original_line = $original_line;
					$exception_block = 0;

					if ($rest =~ /^[\w\.]*\.\w+Exception:/) {
						# This is for the case of two exception blocks one after the other
						$exception_block = 1;
					}
				}
			}
		} 
		
	} else {
		# probably a stack trace so merge with previous log lines
		die "Why I am here? [$line]\n";
	}

	$line_num++;
}

close LOG_INPUT;

# make sure it's all dumped
if ($prev_line ne "") {
	$TMP_HASH{$line_num} = $prev_header . $prev_line;
	$prev_line = "";
	$prev_header = "";

	if ($prev_line =~ /\t\t\t\t\t/ and $exception_block == 1) {
		$EXCEPTION_LINE_HASH{$line_num}++;
	} else {
		$ORIGINAL_LINE_HASH{$line_num} = $prev_original_line;
	}
}

# Now we are going to do log abstraction
while (($name, $value) = (each %TMP_HASH)) {
#	print STDERR "1111 $name $value\n";
	$sample_line = $value;

	######## skip the exception line block, it should not be annomized
	if (exists $EXCEPTION_LINE_HASH{$name}) {
		if (not exists $EVENT_TYPE_HASH{$value}) {
			$EVENT_TYPE_HASH{$value} = $value;
		}
		
		$EVENT_MAP_HASH{$name} = $value;
		next;
	}

	# step 2. anomanize the log line
	if ($value =~ /\b0x[\da-f]+\b/i) {
		# abstract hex numbers 
		# I/ActivityManager(  386): START u0 {flg=0x10000000 cmp=com.android2.calculator3/.Calculator} from pid 29068
		$value =~ s/\b0x[\da-f]+\b/0/g;
		# print STDERR "3456 $value\n";
	}
	if ($value =~ /\@[\da-f]{4}/) {
		# abstract the memory addresses in lines like 
		# I/AudioService(  386):  AudioFocus  abandonAudioFocus() from android.media.AudioManager@41e30a88org.videolan.vlc.AudioService$1@41e30a78
		# these tokens are annoying because they abut other tokens
		$value =~ s/\@[\da-f]{8}/\@MEMOADDR/g;
	}
	if ($value =~ /\w+=\d+/) {
		# abstract the name=value pairs
		# I/ActivityManager(  386): Start proc com.android2.calculator3 for activity com.android2.calculator3/.Calculator: pid=29078 uid=10062 gids={50062, 1028}
		$value =~ s/(\w+)=\d+/$1=0/g;
		# print STDERR "7890 $value\n";
	}

	if ($value =~ /\w+=\{\d+[,\s\d]*\}/) {
		# abstract the name={value1,value2,value3,...} pairs
		# I/ActivityManager(  386): Start proc com.android2.calculator3 for activity com.android2.calculator3/.Calculator: pid=29078 uid=10062 gids={50062, 1028}
		$value =~ s/(\w+)=\{\d+[\,\s\d]*\}/$1=0/g;
		# print STDERR "1122 $value\n";
	}

	# print STDERR "**** After anomanize: $value\n";

	# step 3. tokenize words
	@tmp_array = split(/\W+/, $value);
	$tmp_string = "";

	for ($index=0; $index<=$#tmp_array; $index++) {
		if ($tmp_array[$index] =~ /^\d+$/ 
			or $tmp_array[$index] =~ /^\d+k$/i or $tmp_array[$index] =~ /^\d+m$/i 
			or $tmp_array[$index] =~ /^\d+kb$/i or $tmp_array[$index] =~ /^\d+mb$/i 
			or $tmp_array[$index] =~ /^\d+ms$/ 
			or $tmp_array[$index] =~ /^\d+x\d+$/ 
			or length($tmp_array[$index]) == 1) {
			# remove all the words which are only numbers, or like 12k or 12K, or 12m, or 12M or 12ms
			# GC_CONCURRENT freed 177K, 3% free 8938K/9160K, paused 4ms+3ms, total 18ms
			# core video output: original format sz 320x192, of (0,0), vsz 320x180, 4cc I420, sar 1:1, msk r0x0 g0x0 b0x0
			$word = 0;
		} elsif (length($tmp_array[$index]) == 0) {
			$word = "";
		} else {
			$word = $tmp_array[$index];
		}

		if ($index > 0) {
			$tmp_string = $tmp_string . " " . $word;
		} else {
			$tmp_string = $word;
		}
	}

	# provide an example of the event type hash
	if (not exists $EVENT_TYPE_HASH{$tmp_string}) {
		
		if (not exists $ORIGINAL_LINE_HASH{$name}) {
			die "not existed: $name, $value\n";
		}
		
		$original_line = $ORIGINAL_LINE_HASH{$name};
		$EVENT_TYPE_HASH{$tmp_string} = $original_line;
	}

	$EVENT_MAP_HASH{$name} = $tmp_string;
	# print STDERR "**** After tokenize: $tmp_string\n";

	# step 4. extract the id values
	# removed for logcat, as it will not be used

	# step 5. output both the parameter values as well as event types
	# removed for logcat, as it will not be used
}

my $base_file = basename($file_name);
$output_file = "event_type_" . $base_file . ".dat";

open(EVENT_TYPE_OUTPUT, ">", $output_file) or die "Cannot open event_type.dat: $!\n";

while (($name, $sample_line) = (each %EVENT_TYPE_HASH)) {
	print EVENT_TYPE_OUTPUT "$name||||||||||$sample_line||||||||||$base_file\n";
}

close EVENT_TYPE_OUTPUT;

$output_file = "event_map_" . $base_file . ".dat";

if ($output_file =~ /\//) {
	$output_file =~ s/\//___/g;
#die "My file is: $file_name $output_file\n";
}

open EVENT_MAP_OUTPUT, ">$output_file" or die "Cannot open event_map $output_file: $!\n";

while (($line_num, $event_id) = (each %EVENT_MAP_HASH)) {
	print EVENT_MAP_OUTPUT "$line_num||||||||||$event_id||||||||||$base_file\n";
}

close EVENT_MAP_OUTPUT;

use JSON;

$output_file = "event_count_" . $base_file . ".json";

open EVENT_COUNT_OUTPUT, ">$output_file" or die "Cannot open event_map $output_file: $!\n";

my %count = ();
while (($line_num, $event_id) = (each %EVENT_MAP_HASH)) {
	$count{$event_id}++;
}
print EVENT_COUNT_OUTPUT to_json(\%count, {utf8 => 1, pretty => 1});
close EVENT_COUNT_OUTPUT;
